
/******************************************************************************
* MODULE     : function.hpp
* DESCRIPTION: Functions
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_FUNCTION_HPP
#define __MML_FUNCTION_HPP
#include "string.hpp"
#include "table.hpp"
namespace mml {
#define TMPL_DEF template<typename C, typename T>
#define TMPL template<typename C, typename T>
#define Function function<C,T>
#define Function_rep function_rep<C,T>
TMPL class function_rep;
TMPL class function;

/******************************************************************************
* Functions
******************************************************************************/

TMPL_DEF
class function_rep REP_STRUCT {
protected:
  bool       flag;  // remember previously computed values?
  table<C,T> t;     // remembered values

public:
  Function_rep (): flag (false), t () {}
  Function_rep (const table<C,T>& t2): flag (false), t (t2) {}

  virtual ~Function_rep () {}
  virtual C eval (const T& x) = 0;
  virtual ostream& print_lisp (ostream& out) = 0;

  C var_eval (const T& x) {
    if (!t->contains (x))
      t (x)= eval (x);
    return t [x];
  }
  void remember (bool flag2) {
    flag= flag2;
    if (!flag) t= table<C,T> (I(t));
  }

  friend class Function;
};

TMPL_DEF
class function {
INDIRECT_PROTO_2 (function, function_rep, C, T)
  function ();
  function (int i);
  function (const C& init);
  function (const table<C,T>& t);
  function (C (*fun) (T), string name= "explicit");
  C operator () (const T& x) {
    return rep->flag? rep->var_eval (x): rep->eval (x); }
};
INDIRECT_IMPL_2 (function, function_rep, typename C, C, typename T, T)

TMPL ostream& print_lisp (ostream& out, const Function& f) {
  return f->print_lisp (out); }
TMPL ostream& operator << (ostream& out, const Function& f) {
  return f->print_lisp (out); }

/******************************************************************************
* Constant functions
******************************************************************************/

TMPL_DEF
class constant_function_rep: public Function_rep {
  C c;
public:
  constant_function_rep (const C& c2): c (c2) {}
  C eval (const T& t) { return c; }
  ostream& print_lisp (ostream& out) { return mml::print_lisp (out, c); }
};

TMPL Function::function () {
  rep= new constant_function_rep<C,T> (C ()); }
TMPL Function::function (int c) {
  rep= new constant_function_rep<C,T> (C (c)); }
TMPL Function::function (const C& c) {
  rep= new constant_function_rep<C,T> (c); }

/******************************************************************************
* Explicit functions
******************************************************************************/

TMPL_DEF
class table_function_rep: public Function_rep {
  table<C,T> u;
public:
  table_function_rep (const table<C,T>& u2): u (u2) {}
  C eval (const T& x) { return u[x]; }
  ostream& print_lisp (ostream& out) { return mml::print_lisp (out, u); }
};

TMPL Function::function (const table<C,T>& t) {
  rep= new table_function_rep<C,T> (t); }

TMPL_DEF
class cpp_function_rep: public Function_rep {
  C (*fun) (T);
  string name;
public:
  cpp_function_rep (C (*fun2) (T), string name2): fun (fun2), name (name2) {}
  C eval (const T& x) { return fun (x); }
  ostream& print_lisp (ostream& out) { return mml::print_lisp (out, name); }
};

TMPL Function::function (C (*fun) (T), string name) {
  rep= new cpp_function_rep<C,T> (fun, name); }

/******************************************************************************
* Vectorial arithmetic on functions
******************************************************************************/

#define FUNCTION_BINARY(binop_function_rep,op,disp)		\
TMPL_DEF							\
class binop_function_rep: public Function_rep {			\
  Function f, g;						\
public:								\
  binop_function_rep (const Function& f2, const Function& g2):	\
    f (f2), g (g2) {}						\
  C eval (const T& x) { return f(x) op g(x); }			\
  ostream& print_lisp (ostream& out) {				\
    return mml::print_lisp (out, disp, f, g); }			\
};								\
								\
TMPL Function							\
operator op (const Function& f, const Function& g) {		\
  return Function (new binop_function_rep<C,T> (f, g), false);	\
}

FUNCTION_BINARY(plus_function_rep,+,"+");
FUNCTION_BINARY(minus_function_rep,-,"-");
FUNCTION_BINARY(times_function_rep,*,"*");
FUNCTION_BINARY(over_function_rep,/,"/");
FUNCTION_BINARY(remainder_function_rep,%,"%");
ARITH_INT_SUGAR(TMPL,Function,C);
ARITH_SCALAR_SUGAR(TMPL,Function,C);

#undef TMPL_DEF
#undef TMPL
#undef Function
#undef Function_rep
} // namespace mml
#endif // __MML_FUNCTION_HPP
