
/******************************************************************************
* MODULE     : string.hpp
* DESCRIPTION: Strings with possible zero-characters
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __STRING_HPP
#define __STRING_HPP
#include "base.hpp"
namespace mml {

class string_rep REP_STRUCT {
private:
  char* a;  // the string
  nat   n;  // length of string
  nat   l;  // allocated number of bytes

public:
  inline string_rep (nat n2): a(new char[n2]), n(n2), l(n2) {}
  inline ~string_rep () { delete[] a; }
  void extend (nat n);

  friend class string;
  friend inline nat N (const string& s);
  friend inline char* S (const string& s);
};

class string {
INDIRECT_PROTO (string, string_rep)
public:
  inline string (nat n=0): rep (new string_rep (n)) {}
  string (char c);
  string (char *s);
  string (char *s, nat n);

  inline friend nat N (const string& s) { return s.rep->n; }
  inline friend char* S (const string& s) { return s.rep->a; }
  inline char operator [] (nat i) const { return rep->a[i]; }
  inline char& operator [] (nat i) { return rep->a[i]; }
  string operator () (nat start, nat end);
  bool operator == (char* s);
  bool operator != (char* s);
  bool operator == (const string& s);
  bool operator != (const string& s);
  friend string copy (const string& s);
  friend ostream& operator << (ostream& out, const string& s);

  friend class string_rep;
};
INDIRECT_IMPL (string, string_rep)

string& operator << (string& s, char c);
string& operator << (string& s, const string& t);
string  operator *  (const string& s, const string& t);
bool    operator <  (const string& s, const string& t);
bool    operator <= (const string& s, const string& t);
bool    operator >  (const string& s, const string& t);
bool    operator >= (const string& s, const string& t);

} // namespace mml
#endif // __STRING_HPP
